"""
Functions for basic processing of OMorFi segment labelling style analyses.
"""
import re
from more_itertools import split_at

LABELSEGMENT_RE = r"""
    \{ (?P<seg> [^\}]* ) \} |
    \[ (?P<tag> [^\]]* ) \] |
    (?P<surf> [^\[\{]+ )
"""

_labelsegment_lex = None


def get_labelsegment_lex():
    global _labelsegment_lex
    if _labelsegment_lex is None:
        _labelsegment_lex = re.compile(LABELSEGMENT_RE, re.VERBOSE)
    return _labelsegment_lex


def labelsegment_to_tokens(labelsegmented):
    lex = get_labelsegment_lex()
    for match in re.finditer(lex, labelsegmented):
        typ = match.lastgroup
        value = match.group(typ)
        yield typ, value


def tokens_to_subword_tokens(it):

    def is_cmp_bound(kv):
        return (kv[0] == "seg" and kv[1] == "wB")

    return split_at(it, is_cmp_bound)


def tokens_to_surf(it):
    """
    Given an iterator of segments as (type, value) tuples, reconstruct the
    surface string.
    """
    return "".join(v for (t, v) in it if t == "surf")


def labelsegment_to_subword_tokens(labelsegmented):
    """
    Returns a iterator of segments specified as (type, value) tuple. Type is
    one of "seg", "tag" or "surf".
    """
    return tokens_to_subword_tokens(labelsegment_to_tokens(labelsegmented))



from .omor.tok import get_token_positions
from .omor.inst import get_omorfi
from .omor.anlys import analysis_to_subword_dicts

from .omor.extract import extract_lemmas, extract_lemmas_combs, extract_lemmas_recurs

__all__ = [
    "get_token_positions",
    "get_omorfi",
    "analysis_to_subword_dicts",
    "extract_lemmas",
    "extract_lemmas_combs",
    "extract_lemmas_recurs",
]



from nltk.tokenize import word_tokenize
from nltk.corpus import wordnet
from finntk.emb.base import BothVectorSpaceAdapter, MonoVectorSpaceAdapter
from finntk.emb.utils import apply_vec
from finntk.vendor.metanl.nltk_morphy import tag_and_stem
from finntk.wordnet.reader import fiwn
from itertools import repeat


def lemmatize_tokens(tokens):
    # TODO: Consider switching to ConceptNet5 or Stanford taggger for lemmatization
    # TODO: Consider not doing lemmatization at all and just relying on glosswordnet
    return (stem for (stem, _, _) in tag_and_stem(tokens))


def unexpanded_defn_getter(lemma):
    return word_tokenize(lemma.synset().definition())


def expanded_defn_getter(lemma):
    lemma_name = lemma.name()
    synset = lemma.synset()
    pos = synset.pos()
    if pos == "n":
        schedule = ["Hypernym", "Hyponym", "Holonym", "Meronym", "Attribute"]
    elif pos == "v":
        schedule = ["Hypernym", "Hyponym", "Also see"]
    elif pos in ("s", "a"):
        schedule = ["Attribute", "Also see", "Similar to", "Pertainym of"]
    else:
        schedule = []
    related_synsets = {synset}
    for relation in schedule:
        if relation == "Hypernym":
            related_synsets.update(synset.hypernyms())
            related_synsets.update(synset.instance_hypernyms())
        elif relation == "Hyponym":
            related_synsets.update(synset.hyponyms())
            related_synsets.update(synset.instance_hyponyms())
        elif relation == "Holonym":
            related_synsets.update(synset.member_holonyms())
            related_synsets.update(synset.substance_holonyms())
            related_synsets.update(synset.part_holonyms())
        elif relation == "Meronym":
            related_synsets.update(synset.member_meronyms())
            related_synsets.update(synset.substance_meronyms())
            related_synsets.update(synset.part_meronyms())
        elif relation == "Attribute":
            related_synsets.update(synset.attributes())
        elif relation == "Also see":
            related_synsets.update(synset.also_sees())
        elif relation == "Similar to":
            related_synsets.update(synset.similar_tos())
        elif relation == "Pertainym of":
            related_synsets.update((l.synset() for l in lemma.pertainyms()))
        else:
            assert False, "Unknown relation"
    tokens = []
    for synset in related_synsets:
        tokens.extend(word_tokenize(synset.definition()))
        tokens.extend(
            (lemma.name() for lemma in synset.lemmas() if lemma.name() != lemma_name)
        )
    return tokens


def safe_cosine_sim(u, v):
    from finntk.emb.utils import cosine_sim

    if u is None or v is None:
        return -2
    return cosine_sim(u, v)


def disambg_one(lemma_defns, context_vec, freqs=None):
    if freqs is None:
        freqs = repeat(None)
    best_lemma = None
    best_sim = float("-inf")
    if context_vec is None:
        return best_lemma, best_sim
    for (lemma, defn_vec), freq in zip(lemma_defns, freqs):
        if defn_vec is None:
            continue
        sim = safe_cosine_sim(context_vec, defn_vec)
        if freq is not None:
            if sim >= 0:
                sim *= freq
            else:
                sim /= freq
        if sim > best_sim:
            best_lemma = lemma
            best_sim = sim
    return best_lemma, best_sim


def wn_filter_stream(wn, stream):
    return ((to, le) for to, le in stream if len(wn.lemmas(le)))


class MultilingualLesk:

    def __init__(self, multispace, aggf, wn_filter=False, expand=False, use_freq=False):
        self.multispace = multispace
        self.defn_space = BothVectorSpaceAdapter(
            MonoVectorSpaceAdapter(multispace, "en")
        )
        self.ctx_space = BothVectorSpaceAdapter(
            MonoVectorSpaceAdapter(multispace, "fi")
        )
        self.aggf = aggf
        self.wn_filter = wn_filter
        self.expand = expand
        self.use_freq = use_freq

    def mk_defn_vec(self, item):
        if self.expand:
            defn_tokens = expanded_defn_getter(item)
        else:
            defn_tokens = unexpanded_defn_getter(item)

        lemmatized = lemmatize_tokens(defn_tokens)
        stream = zip(defn_tokens, lemmatized)
        if self.wn_filter:
            stream = wn_filter_stream(wordnet, stream)

        vec = apply_vec(self.aggf, self.defn_space, stream, "en")
        return vec

    def mk_ctx_vec(self, context):
        if self.wn_filter:
            context = wn_filter_stream(fiwn, context)
        vec = apply_vec(self.aggf, self.ctx_space, context, "fi")
        return vec

    def disambg_one(self, choice_freqs, context):
        if self.use_freq:
            choices = []
            freqs = []
            for choice, freq in choice_freqs:
                choices.append(choice)
                freqs.append(freq)
        else:
            choices = choice_freqs
            freqs = None
        return disambg_one(
            ((item, self.mk_defn_vec(item)) for item in choices),
            self.mk_ctx_vec(context),
            freqs,
        )



import gzip
import logging
from finntk.utils import ResourceMan, urlretrieve
from finntk.vendor.conceptnet5.uri import concept_uri
from gensim.models import KeyedVectors
from shutil import copyfileobj
import os
from .base import MultilingualVectorSpace, RefType
from .utils import get

logger = logging.getLogger(__name__)


class NumberbatchWordVecs(ResourceMan):
    RESOURCE_NAME = "numberbatch-multilingual"

    URL = (
        "https://conceptnet.s3.amazonaws.com/downloads/2017/numberbatch"
        "/numberbatch-17.06.txt.gz"
    )

    def __init__(self):
        super().__init__()
        self._resources["vecs"] = "numberbatch.multi.binvec"
        self._vecs = None

    def _bootstrap(self, _res):
        from gensim.test.utils import get_tmpfile

        logger.info("Downloading word vectors")
        gzipped_glove_tmp_fn = urlretrieve(self.URL)
        try:
            glove_tmp_fn = get_tmpfile("glove.txt")
            try:
                copyfileobj(gzip.open(gzipped_glove_tmp_fn), open(glove_tmp_fn, "wb"))
                logger.info("Converting word vectors")
                fi = KeyedVectors.load_word2vec_format(glove_tmp_fn)
                fi.save(self._get_res_path("vecs"))
            finally:
                try:
                    os.remove(glove_tmp_fn)
                except OSError:
                    pass
        finally:
            try:
                os.remove(gzipped_glove_tmp_fn)
            except OSError:
                pass

    def get_vecs(self):
        if self._vecs is None:
            vec_path = self.get_res("vecs")
            logger.info("Loading word vectors")
            self._vecs = KeyedVectors.load(vec_path, mmap="r")
            logger.info("Loaded word vectors")
        return self._vecs


vecs = NumberbatchWordVecs()


def mk_concept_vec(lang, text, *more):
    return get(vecs.get_vecs(), concept_uri(lang, text, *more))


class NumberbatchMultiSpace(MultilingualVectorSpace):
    takes = RefType.LEMMA
    dim = 300

    def get_vec(self, lang: str, ref: str):
        return mk_concept_vec(lang, ref)


multispace = NumberbatchMultiSpace()



from abc import ABC, abstractmethod
from enum import Enum
from typing import Tuple


class RefType(Enum):
    WORD_FORM = 1
    LEMMA = 2
    BOTH = 3


class LexicalVectorSpace(ABC):
    takes: RefType
    dim: int


class MultilingualVectorSpace(LexicalVectorSpace):

    @abstractmethod
    def get_vec(self, lang: str, ref):
        pass


class MonolingualVectorSpace(LexicalVectorSpace):

    @abstractmethod
    def get_vec(self, ref):
        pass


class MonoVectorSpaceAdapter(MonolingualVectorSpace):
    lang: str

    def __init__(self, inner: MultilingualVectorSpace, lang: str):
        self.inner = inner
        self.lang = lang
        self.takes = self.inner.takes
        self.dim = self.inner.dim

    def get_vec(self, ref):
        return self.inner.get_vec(self.lang, ref)


class BothVectorSpaceAdapter(MonolingualVectorSpace):
    takes = RefType.BOTH

    def __init__(self, inner: MonolingualVectorSpace):
        self.inner = inner
        self.dim = self.inner.dim

    def get_vec(self, ref: Tuple[str, str]):
        if self.inner.takes == RefType.WORD_FORM:
            return self.inner.get_vec(ref[0])
        elif self.inner.takes == RefType.LEMMA:
            return self.inner.get_vec(ref[1])
        else:
            return self.inner.get_vec(ref)



# -*- coding: utf-8 -*-
from __future__ import print_function, unicode_literals

import nltk
from nltk.corpus import wordnet

try:
    morphy = wordnet._morphy
except LookupError:
    nltk.download('wordnet')
    morphy = wordnet._morphy

STOPWORDS = ['the', 'a', 'an']

EXCEPTIONS = {
    # Avoid obsolete and obscure roots, the way lexicographers don't.
    'wrought': 'wrought',   # not 'work'
    'media': 'media',       # not 'medium'
    'installed': 'install', # not 'instal'
    'installing': 'install',# not 'instal'
    'synapses': 'synapse',  # not 'synapsis'
    'soles': 'sole',        # not 'sol'
    'pubes': 'pube',        # not 'pubis'
    'dui': 'dui',           # not 'duo'
    'taxis': 'taxi',        # not 'taxis'

    # Work around errors that Morphy makes.
    'alas': 'alas',
    'corps': 'corps',
    'cos': 'cos',
    'enured': 'enure',
    'fiver': 'fiver',
    'hinder': 'hinder',
    'lobed': 'lobe',
    'offerer': 'offerer',
    'outer': 'outer',
    'sang': 'sing',
    'singing': 'sing',
    'solderer': 'solderer',
    'tined': 'tine',
    'twiner': 'twiner',
    'us': 'us',

    # Stem common nouns whose plurals are apparently ambiguous
    'teeth': 'tooth',
    'things': 'thing',
    'people': 'person',

    # Tokenization artifacts
    'wo': 'will',
    'ca': 'can',
    "n't": 'not',
}

AMBIGUOUS_EXCEPTIONS = {
    # Avoid nouns that shadow more common verbs.
    'am': 'be',
    'as': 'as',
    'are': 'be',
    'ate': 'eat',
    'bent': 'bend',
    'drove': 'drive',
    'fell': 'fall',
    'felt': 'feel',
    'found': 'find',
    'has': 'have',
    'lit': 'light',
    'lost': 'lose',
    'sat': 'sit',
    'saw': 'see',
    'sent': 'send',
    'shook': 'shake',
    'shot': 'shoot',
    'slain': 'slay',
    'spoke': 'speak',
    'stole': 'steal',
    'sung': 'sing',
    'thought': 'think',
    'tore': 'tear',
    'was': 'be',
    'won': 'win',
    'feed': 'feed',
}


def _word_badness(word):
    """
    Assign a heuristic to possible outputs from Morphy. Minimizing this
    heuristic avoids incorrect stems.
    """
    if word.endswith('e'):
        return len(word) - 2
    elif word.endswith('ess'):
        return len(word) - 10
    elif word.endswith('ss'):
        return len(word) - 4
    else:
        return len(word)


def _morphy_best(word, pos=None):
    """
    Get the most likely stem for a word using Morphy, once the input has been
    pre-processed by morphy_stem().
    """
    results = []
    if pos is None:
        pos = 'nvar'
    for pos_item in pos:
        results.extend(morphy(word, pos_item))
    if not results:
        return None
    results.sort(key=lambda x: _word_badness(x))
    return results[0]


def morphy_stem(word, pos=None):
    """
    Get the most likely stem for a word. If a part of speech is supplied,
    the stem will be more accurate.
    Valid parts of speech are:
    - 'n' or 'NN' for nouns
    - 'v' or 'VB' for verbs
    - 'a' or 'JJ' for adjectives
    - 'r' or 'RB' for adverbs
    Any other part of speech will be treated as unknown.
    """
    word = word.lower()
    if pos is not None:
        if pos.startswith('NN'):
            pos = 'n'
        elif pos.startswith('VB'):
            pos = 'v'
        elif pos.startswith('JJ'):
            pos = 'a'
        elif pos.startswith('RB'):
            pos = 'r'
    if pos is None and word.endswith('ing') or word.endswith('ed'):
        pos = 'v'
    if pos is not None and pos not in 'nvar':
        pos = None
    if word in EXCEPTIONS:
        return EXCEPTIONS[word]
    if pos is None:
        if word in AMBIGUOUS_EXCEPTIONS:
            return AMBIGUOUS_EXCEPTIONS[word]
    return _morphy_best(word, pos) or word


def tag_and_stem(tokens):
    """
    Returns a list of (stem, tag, token) triples:
    - stem: the word's uninflected form
    - tag: the word's part of speech
    - token: the original word, so we can reconstruct it later
    """
    tagged = nltk.pos_tag(tokens)
    out = []
    for token, tag in tagged:
        stem = morphy_stem(token, tag)
        out.append((stem, tag, token))
    return out


def good_lemma(lemma):
    return lemma and lemma not in STOPWORDS and lemma[0].isalnum()


def normalize_list(tokens):
    """
    Get a list of word stems that appear in the text. Stopwords and an initial
    'to' will be stripped.
    """
    pieces = [morphy_stem(word) for word in tokens]
    pieces = [piece for piece in pieces if good_lemma(piece)]
    if not pieces:
        return []
    if pieces[0] == 'to':
        pieces = pieces[1:]
    return pieces



from finntk.utils import ResourceMan, LazyCorpusLoader
from os.path import exists, join as pjoin
from nltk.corpus import wordnet
from nltk.corpus.reader.wordnet import (
    WordNetCorpusReader,
    Synset,
    Lemma,
    VERB_FRAME_STRINGS,
    ADJ,
    ADJ_SAT,
    WordNetError,
)
import logging
import re
from plumbum.cmd import git
import portalocker

logger = logging.getLogger(__name__)


# This lemma is broken in data.noun
LEMMA_NAME_FIXES = {"moderniusAdditional_synonym": "modernius"}

# These lemmas occur in differently upper/lower case wise in transls and data.POS
IGNORE_CASE_LEMMAS = {"ci", "otc", "rh-negatiivinen", "rh-positiivinen"}


class FinnWordNetResMan(ResourceMan):
    RESOURCE_NAME = "fiwn"
    RELS = ["transls"]

    REPO = "https://github.com/frankier/fiwn.git"

    def __init__(self):
        super().__init__()
        for name in self.RELS:
            self._resources[name] = pjoin("data", "rels", "fiwn-{}.tsv".format(name))
        self._resources["synset_map"] = "synset_map.tsv"

    def _get_res_filename(self, res):
        return self._resources.get(res, pjoin("data", "dict", res))

    @property
    def _done_file(self):
        data_dir = self._get_data_dir()
        return pjoin(data_dir, ".done")

    def _is_bootstrapped(self):
        return exists(self._done_file)

    def _bootstrap(self, _res=None):
        data_dir = self._get_data_dir()
        with portalocker.Lock(pjoin(data_dir, "../fiwn.lock"), timeout=3600):
            if not self._is_bootstrapped():
                git("clone", self.REPO, data_dir)
                with open(self._done_file, "w"):
                    pass


fiwn_resman = FinnWordNetResMan()


class FinnWordNetReader(WordNetCorpusReader):

    def __init__(self, *args, **kwargs):
        fiwn_resman.bootstrap()
        args = list(args)
        args[0] = fiwn_resman.get_res("")
        super().__init__(*args, **kwargs)

    def open(self, filename):
        if filename == "lexnames":
            return wordnet.open("lexnames")
        return super().open(filename)

    def _synset_from_pos_and_line(self, pos, data_file_line):
        # Copied and modified for FiWN. The only change is the handling of
        # FiWN's special <tag/> syntactic markers.

        # Construct a new (empty) synset.
        synset = Synset(self)

        # parse the entry for this synset
        try:

            # parse out the definitions and examples from the gloss
            columns_str, gloss = data_file_line.split("|")
            gloss = gloss.strip()
            definitions = []
            for gloss_part in gloss.split(";"):
                gloss_part = gloss_part.strip()
                if gloss_part.startswith('"'):
                    synset._examples.append(gloss_part.strip('"'))
                else:
                    definitions.append(gloss_part)
            synset._definition = "; ".join(definitions)

            # split the other info into fields
            _iter = iter(columns_str.split())

            def _next_token():
                return next(_iter)

            # get the offset
            synset._offset = int(_next_token())

            # determine the lexicographer file name
            lexname_index = int(_next_token())
            synset._lexname = self._lexnames[lexname_index]

            # get the part of speech
            synset._pos = _next_token()

            # create Lemma objects for each lemma
            n_lemmas = int(_next_token(), 16)
            for _ in range(n_lemmas):
                # get the lemma name
                lemma_name = _next_token()
                # get the lex_id (used for sense_keys)
                lex_id = int(_next_token(), 16)
                # If the lemma has a syntactic marker, extract it.
                m = re.match(r"([^<]*)(<.*)?$", lemma_name)
                lemma_name, syn_mark = m.groups()
                # create the lemma object
                lemma = Lemma(self, synset, lemma_name, lexname_index, lex_id, syn_mark)
                synset._lemmas.append(lemma)
                synset._lemma_names.append(lemma._name)

            # collect the pointer tuples
            n_pointers = int(_next_token())
            for _ in range(n_pointers):
                symbol = _next_token()
                offset = int(_next_token())
                pos = _next_token()
                lemma_ids_str = _next_token()
                if lemma_ids_str == "0000":
                    synset._pointers[symbol].add((pos, offset))
                else:
                    source_index = int(lemma_ids_str[:2], 16) - 1
                    target_index = int(lemma_ids_str[2:], 16) - 1
                    source_lemma_name = synset._lemmas[source_index]._name
                    lemma_pointers = synset._lemma_pointers
                    tups = lemma_pointers[source_lemma_name, symbol]
                    tups.append((pos, offset, target_index))

            # read the verb frames
            try:
                frame_count = int(_next_token())
            except StopIteration:
                pass
            else:
                for _ in range(frame_count):
                    # read the plus sign
                    plus = _next_token()
                    assert plus == "+"
                    # read the frame and lemma number
                    frame_number = int(_next_token())
                    frame_string_fmt = VERB_FRAME_STRINGS[frame_number]
                    lemma_number = int(_next_token(), 16)
                    # lemma number of 00 means all words in the synset
                    if lemma_number == 0:
                        synset._frame_ids.append(frame_number)
                        for lemma in synset._lemmas:
                            lemma._frame_ids.append(frame_number)
                            lemma._frame_strings.append(frame_string_fmt % lemma._name)
                    # only a specific word in the synset
                    else:
                        lemma = synset._lemmas[lemma_number - 1]
                        lemma._frame_ids.append(frame_number)
                        lemma._frame_strings.append(frame_string_fmt % lemma._name)

        # raise a more informative error with line text
        except ValueError as e:
            raise WordNetError("line %r: %s" % (data_file_line, e))

        # set sense keys for Lemma objects - note that this has to be
        # done afterwards so that the relations are available
        for lemma in synset._lemmas:
            if synset._pos == ADJ_SAT:
                head_lemma = synset.similar_tos()[0]._lemmas[0]
                head_name = head_lemma._name
                head_id = "%02d" % head_lemma._lex_id
            else:
                head_name = head_id = ""
            tup = (
                lemma._name,
                WordNetCorpusReader._pos_numbers[synset._pos],
                lemma._lexname_index,
                lemma._lex_id,
                head_name,
                head_id,
            )
            lemma._key = ("%s%%%d:%02d:%02d:%s:%s" % tup).lower()

        # the canonical name is based on the first lemma
        lemma_name = synset._lemmas[0]._name.lower()
        offsets = self._lemma_pos_offset_map[lemma_name][synset._pos]
        sense_index = offsets.index(synset._offset)
        tup = lemma_name, synset._pos, sense_index + 1
        synset._name = "%s.%s.%02i" % tup

        return synset

    def _load_lemma_pos_offset_map(self):
        # Copied and modified for FiWN. The only changes are to handle a single
        # entry: FiWN contains an empty lemma for `taken` as in `taken
        # ill/drunk`. Thus lines beginning with two spaces as are treated as a
        # comment instead of ones starting with one space and tokenisation is
        # done on spaces instead of all whitespace.

        for suffix in self._FILEMAP.values():

            # parse each line of the file (ignoring comment lines)
            for i, line in enumerate(self.open("index.%s" % suffix)):
                if line.startswith("  "):
                    continue

                _iter = iter(line.split(" "))

                def _next_token():
                    return next(_iter)

                try:

                    # get the lemma and part-of-speech
                    lemma = _next_token()
                    pos = _next_token()

                    # get the number of synsets for this lemma
                    n_synsets = int(_next_token())
                    assert n_synsets > 0

                    # get and ignore the pointer symbols for all synsets of
                    # this lemma
                    n_pointers = int(_next_token())
                    [_next_token() for _ in range(n_pointers)]

                    # same as number of synsets
                    n_senses = int(_next_token())
                    assert n_synsets == n_senses

                    # get and ignore number of senses ranked according to
                    # frequency
                    _next_token()

                    # get synset offsets
                    synset_offsets = [int(_next_token()) for _ in range(n_synsets)]

                # raise more informative error with file name and line number
                except (AssertionError, ValueError) as e:
                    tup = ("index.%s" % suffix), (i + 1), e
                    raise WordNetError("file %s, line %i: %s" % tup)

                # map lemmas and parts of speech to synsets
                self._lemma_pos_offset_map[lemma][pos] = synset_offsets
                if pos == ADJ:
                    self._lemma_pos_offset_map[lemma][ADJ_SAT] = synset_offsets


fiwn = LazyCorpusLoader("fiwn", FinnWordNetReader, None)


class CountCachingMixin:

    def __init__(self, *args, **kwargs) -> None:
        self._count_cache = {}
        super().__init__(*args, **kwargs)

    def lemma_count(self, lemma: Lemma) -> int:
        if lemma._key not in self._count_cache:
            self._count_cache[lemma._key] = super().lemma_count(lemma)
        return self._count_cache[lemma._key]


class LemmaFreqMixin:
    ADD_CONST = 1000

    def lemma_freqs(self, lemma: str, pos=None, lang="eng") -> float:
        lemmas = self.lemmas(lemma, pos, lang)
        adjusted_counts = [lemma.count() + self.ADD_CONST for lemma in lemmas]
        total = sum(adjusted_counts)
        return zip(lemmas, (adj_count / total for adj_count in adjusted_counts))


class EnCountsFinnWordNetReader(CountCachingMixin, LemmaFreqMixin, FinnWordNetReader):

    def lemmas(self, lemma, pos=None, lang="eng"):
        lemmas = super().lemmas(lemma, pos, lang)
        return sorted(lemmas, key=lambda l: -l.count())


fiwn_encnt = LazyCorpusLoader("fiwn", EnCountsFinnWordNetReader, None)


en_fi_maps = None


def get_en_fi_maps():

    def fix_synset_key(synset):
        synset = synset.split(":", 1)[1]
        if synset[0] == "s":
            synset = "a" + synset[1:]
        return synset

    global en_fi_maps
    if en_fi_maps is None:
        synset_map = fiwn_resman.get_res("synset_map")
        fi2en = {}
        en2fi = {}
        for line in open(synset_map):
            en_synset_key, fi_synset_key = line[:-1].split("\t")
            en_synset_key = fix_synset_key(en_synset_key)
            fi_synset_key = fix_synset_key(fi_synset_key)
            fi2en[fi_synset_key] = en_synset_key
            en2fi[en_synset_key] = fi_synset_key
        en_fi_maps = fi2en, en2fi
    return en_fi_maps



import logging
from finntk.utils import ResourceMan, urlretrieve
from gensim.models import KeyedVectors
import os
from .base import MultilingualVectorSpace, RefType
from .utils import get

logger = logging.getLogger(__name__)


class FasttextWordVecs(ResourceMan):
    RESOURCE_NAME = "fasttext-multilingual"

    FI_URL = "https://s3.amazonaws.com/arrival/embeddings/wiki.multi.fi.vec"
    EN_URL = "https://s3.amazonaws.com/arrival/embeddings/wiki.multi.en.vec"

    def __init__(self):
        super().__init__()
        self._resources["fi_vec"] = "wiki.multi.fi.binvec"
        self._resources["en_vec"] = "wiki.multi.en.binvec"
        self._en = None
        self._fi = None

    def _download(self, lang, url, dest):
        logger.info("Downloading {} word vectors".format(lang))
        tmp_fn = urlretrieve(url)
        try:
            logger.info("Converting {} word vectors".format(lang))
            fi = KeyedVectors.load_word2vec_format(tmp_fn)
            fi.save(dest)
        finally:
            os.remove(tmp_fn)

    def _bootstrap(self, _res):
        self._download("Finnish", self.FI_URL, self._get_res_path("fi_vec"))
        self._download("English", self.EN_URL, self._get_res_path("en_vec"))

    def get_en(self):
        if self._en is None:
            en_vec_path = self.get_res("en_vec")
            logger.info("Loading English word vectors")
            self._en = KeyedVectors.load(en_vec_path, mmap="r")
            logger.info("Loaded English word vectors")
        return self._en

    def get_fi(self):
        if self._fi is None:
            fi_vec_path = self.get_res("fi_vec")
            logger.info("Loading Finnish word vectors")
            self._fi = KeyedVectors.load(fi_vec_path, mmap="r")
            logger.info("Loaded Finnish word vectors")
        return self._fi


vecs = FasttextWordVecs()


class FasttextMultiSpace(MultilingualVectorSpace):
    takes = RefType.WORD_FORM
    dim = 300

    def get_vec(self, lang: str, ref: str):
        if lang == "en":
            space = vecs.get_en()
        elif lang == "fi":
            space = vecs.get_fi()
        else:
            assert False, "Unsupported language"
        return get(space, ref)


multispace = FasttextMultiSpace()



"""
Functions for basic processing of OMorFi tokens.
"""


def get_token_positions(tokenised, text):
    """
    Returns the start positions of a series of tokens produced by
    Omorfi.tokenise(...)
    """
    starts = []
    start = 0
    for token in tokenised:
        start = text.index(token["surf"], start)
        starts.append(start)
        start += len(token["surf"])
    return starts


def form_of_tok(token):
    if isinstance(token, str):
        return token.lower()
    else:
        return token["surf"].lower()



"""
Functions for extracting lemmas from OMorFi analyses.
"""
from .inst import get_omorfi
from .anlys import analysis_to_subword_dicts, lemmas_of_subword_dicts


def contig_slices(elems):
    elems = list(elems)
    n = len(elems)
    for start in range(n):
        for end in range(start + 1, n + 1):
            yield elems[start:end]


def _extract_lemmas(word_form, get_slices):
    omorfi = get_omorfi()
    analyses = omorfi.analyse(word_form)
    res = set()
    for analysis in analyses:
        analysis_dicts = analysis_to_subword_dicts(analysis["anal"])
        for analysis_slice in get_slices(analysis_dicts):
            for lemma in lemmas_of_subword_dicts(analysis_slice):
                res.add(lemma)
    return res


def extract_lemmas(word_form):
    """
    Extract lemmas specifically mentioned by OMorFi.
    """
    return _extract_lemmas(
        word_form, lambda analysis_dicts: [[d] for d in analysis_dicts]
    )


def extract_lemmas_span(word_form):
    """
    Works like `extract_lemmas`, but doesn't extract individual subwords.
    However, if a word is only recognised by as a compound word by OMorFi it
    will glue the parts together, lemmatising only the last subword. This means
    it extracts only lemmas which span the whole word form.
    """
    return _extract_lemmas(word_form, lambda analysis_dicts: [analysis_dicts])


def extract_lemmas_combs(word_form):
    """
    Works like `extract_lemmas`, but also tries to combine adjacent
    subwords to make lemmas which may be out of volcaburary for
    OMorFi.

    Note that this will over generate (by design). For example:
    voileipäkakku will generate voi, voileipä and voileipäkakku  as
    desired, but will also spuriously generate leipäkakku.
    """
    return _extract_lemmas(word_form, contig_slices)


def extract_lemmas_recurs(word_form):
    """
    Works like `extract_lemmas`, but also tries to expand each
    lemma into more lemmas. This helps in some cases (but can
    overgenerate even more). For example, it will mean that
    synnyinkaupunkini will generate synty, kaupunki,
    synnyinkaupunki, synnyin and syntyä.
    """
    expand_queue = [word_form]
    res = set()
    while len(expand_queue) > 0:
        word_form = expand_queue.pop()
        new_lemmas = extract_lemmas_combs(word_form)
        novel_lemmas = new_lemmas - res
        expand_queue.extend(novel_lemmas)
        for lemma in novel_lemmas:
            res.add(lemma)
    return res


def lemma_intersect(toks1, toks2):
    """
    Given two iterables of tokens, return the intersection of their lemmas.
    This can work as a simple, high recall, method of matching for example, two
    inflected noun phrases.
    """
    if len(toks1) != len(toks2):
        return
    res = []
    for t1, t2 in zip(toks1, toks2):
        l1 = extract_lemmas_span(t1)
        l2 = extract_lemmas_span(t2)
        inter = l1 & l2
        if len(inter) == 0:
            return
        res.append(inter)
    return res



"""
URIs are Unicode strings that represent the canonical name for any object in
ConceptNet. These can be used with the ConceptNet Web API, or referred to in a
Semantic Web application, by attaching the prefix:

    http://api.conceptnet.io

For example, the English concept "book" has the URI '/c/en/book'. This concept
can be referred to, or retrieved, using this complete URI:

    http://api.conceptnet.io/c/en/book
"""


def standardize_text(text, lowercase=True):
    raise NotImplementedError(
        "This function has been superseded by "
        "conceptnet5.nodes.preprocess_and_tokenize_text."
    )


def join_uri(*pieces):
    """
    `join_uri` builds a URI from constituent pieces that should be joined
    with slashes (/).

    Leading and trailing on the pieces are acceptable, but will be ignored.
    The resulting URI will always begin with a slash and have its pieces
    separated by a single slash.

    The pieces do not have `preprocess_and_tokenize_text` applied to them; to make sure your
    URIs are in normal form, run `preprocess_and_tokenize_text` on each piece that represents
    arbitrary text.

    >>> join_uri('/c', 'en', 'cat')
    '/c/en/cat'

    >>> join_uri('c', 'en', ' spaces ')
    '/c/en/ spaces '

    >>> join_uri('/r/', 'AtLocation/')
    '/r/AtLocation'

    >>> join_uri('/test')
    '/test'

    >>> join_uri('test')
    '/test'

    >>> join_uri('/test', '/more/')
    '/test/more'
    """
    joined = "/" + ("/".join([piece.strip("/") for piece in pieces]))
    return joined


def concept_uri(lang, text, *more):
    """
    `concept_uri` builds a representation of a concept, which is a word or
    phrase of a particular language, which can participate in relations with
    other concepts, and may be linked to concepts in other languages.

    Every concept has an ISO language code and a text. It may also have a part
    of speech (pos), which is typically a single letter. If it does, it may
    have a disambiguation, a string that distinguishes it from other concepts
    with the same text.

    This function should be called as follows, where arguments after `text`
    are optional:

        concept_uri(lang, text, pos, disambiguation...)

    `text` and `disambiguation` should be strings that have already been run
    through `preprocess_and_tokenize_text`.

    This is a low-level interface. See `standardized_concept_uri` in nodes.py for
    a more generally applicable function that also deals with special
    per-language handling.

    >>> concept_uri('en', 'cat')
    '/c/en/cat'
    >>> concept_uri('en', 'cat', 'n')
    '/c/en/cat/n'
    >>> concept_uri('en', 'cat', 'n', 'feline')
    '/c/en/cat/n/feline'
    >>> concept_uri('en', 'this is wrong')
    Traceback (most recent call last):
        ...
    AssertionError: 'this is wrong' is not in normalized form
    """
    assert " " not in text, "%r is not in normalized form" % text
    if len(more) > 0:
        if len(more[0]) != 1:
            # We misparsed a part of speech; everything after the text is
            # probably junk
            more = []
        for dis1 in more[1:]:
            assert " " not in dis1, "%r is not in normalized form" % dis1

    return join_uri("/c", lang, text, *more)


def compound_uri(op, args):
    """
    Some URIs represent a compound structure or operator built out of a number
    of arguments. Some examples are the '/and' and '/or' operators, which
    represent a conjunction or disjunction over two or more URIs, which may
    themselves be compound URIs; or the assertion structure, '/a', which takes
    a relation and two URIs as its arguments.

    This function takes the main 'operator', with the slash included, and an
    arbitrary number of arguments, and produces the URI that represents the
    entire compound structure.

    These structures contain square brackets as segments, which look like
    `/[/` and `/]/`, so that compound URIs can contain other compound URIs
    without ambiguity.

    >>> compound_uri('/nothing', [])
    '/nothing/[/]'
    >>> compound_uri('/a', ['/r/CapableOf', '/c/en/cat', '/c/en/sleep'])
    '/a/[/r/CapableOf/,/c/en/cat/,/c/en/sleep/]'
    """
    items = [op]
    first_item = True
    items.append("[")
    for arg in args:
        if first_item:
            first_item = False
        else:
            items.append(",")
        items.append(arg)
    items.append("]")
    return join_uri(*items)


def split_uri(uri):
    """
    Get the slash-delimited pieces of a URI.

    >>> split_uri('/c/en/cat/n/animal')
    ['c', 'en', 'cat', 'n', 'animal']
    >>> split_uri('/')
    []
    """
    if not uri.startswith("/"):
        return [uri]
    uri2 = uri.lstrip("/")
    if not uri2:
        return []
    return uri2.split("/")


def uri_prefix(uri, max_pieces=3):
    """
    Strip off components that might make a ConceptNet URI too detailed. Only
    the first `max_pieces` components will be kept.

    By default, `max_pieces` is 3, making this function useful for converting
    disambiguated concepts into their more general ambiguous forms.

    If the URI is actually a fully qualified URL, no components are removed.

    >>> uri_prefix('/c/en/cat/n/animal')
    '/c/en/cat'
    >>> uri_prefix('/c/en/cat/n')
    '/c/en/cat'
    >>> uri_prefix('/c/en/cat')
    '/c/en/cat'
    >>> uri_prefix('/c/en')
    '/c/en'
    >>> uri_prefix('/c/en/cat', 2)
    '/c/en'
    >>> uri_prefix('http://en.wikipedia.org/wiki/Example')
    'http://en.wikipedia.org/wiki/Example'
    """
    if is_absolute_url(uri):
        return uri
    pieces = split_uri(uri)[:max_pieces]
    return join_uri(*pieces)


def uri_prefixes(uri, min_pieces=2):
    """
    Get URIs that are prefixes of a given URI: that is, they begin with the
    same path components. By default, the prefix must have at least 2
    components.

    If the URI has sub-parts that are grouped by square brackets, then
    only complete sub-parts will be allowed in prefixes.

    >>> list(uri_prefixes('/c/en/cat/n/animal'))
    ['/c/en', '/c/en/cat', '/c/en/cat/n', '/c/en/cat/n/animal']
    >>> list(uri_prefixes('/test/[/group/one/]/[/group/two/]'))
    ['/test/[/group/one/]', '/test/[/group/one/]/[/group/two/]']
    >>> list(uri_prefixes('http://en.wikipedia.org/wiki/Example'))
    ['http://en.wikipedia.org/wiki/Example']
    """
    if is_absolute_url(uri):
        yield uri
        return
    pieces = []
    for piece in split_uri(uri):
        pieces.append(piece)
        if len(pieces) >= min_pieces:
            if pieces.count("[") == pieces.count("]"):
                yield join_uri(*pieces)


def parse_compound_uri(uri):
    """
    Given a compound URI, extract its operator and its list of arguments.

    >>> parse_compound_uri('/nothing/[/]')
    ('/nothing', [])
    >>> parse_compound_uri('/a/[/r/CapableOf/,/c/en/cat/,/c/en/sleep/]')
    ('/a', ['/r/CapableOf', '/c/en/cat', '/c/en/sleep'])
    >>> parse_compound_uri('/or/[/and/[/s/one/,/s/two/]/,/and/[/s/three/,/s/four/]/]')
    ('/or', ['/and/[/s/one/,/s/two/]', '/and/[/s/three/,/s/four/]'])
    """
    pieces = split_uri(uri)
    if pieces[-1] != "]":
        raise ValueError("Compound URIs must end with /]")
    if "[" not in pieces:
        raise ValueError(
            "Compound URIs must contain /[/ at the beginning of " "the argument list"
        )
    list_start = pieces.index("[")
    op = join_uri(*pieces[:list_start])

    chunks = []
    current = []
    depth = 0

    # Split on commas, but not if they're within additional pairs of brackets.
    for piece in pieces[(list_start + 1):-1]:
        if piece == "," and depth == 0:
            chunks.append("/" + ("/".join(current)).strip("/"))
            current = []
        else:
            current.append(piece)
            if piece == "[":
                depth += 1
            elif piece == "]":
                depth -= 1

    assert depth == 0, "Unmatched brackets in %r" % uri
    if current:
        chunks.append("/" + ("/".join(current)).strip("/"))
    return op, chunks


def parse_possible_compound_uri(op, uri):
    """
    The AND and OR conjunctions can be expressed as compound URIs, but if they
    contain only one thing, they are returned as just that single URI, not a
    compound.

    This function returns the list of things in the compound URI if its operator
    matches `op`, or a list containing the URI itself if not.

    >>> parse_possible_compound_uri(
    ...    'or', '/or/[/and/[/s/one/,/s/two/]/,/and/[/s/three/,/s/four/]/]'
    ... )
    ['/and/[/s/one/,/s/two/]', '/and/[/s/three/,/s/four/]']
    >>> parse_possible_compound_uri('or', '/s/contributor/omcs/dev')
    ['/s/contributor/omcs/dev']
    """
    if uri.startswith("/" + op + "/"):
        return parse_compound_uri(uri)[1]
    else:
        return [uri]


def conjunction_uri(*sources):
    """
    Make a URI representing a conjunction of sources that work together to provide
    an assertion. The sources will be sorted in lexicographic order.

    >>> conjunction_uri('/s/contributor/omcs/dev')
    '/s/contributor/omcs/dev'

    >>> conjunction_uri('/s/rule/some_kind_of_parser', '/s/contributor/omcs/dev')
    '/and/[/s/contributor/omcs/dev/,/s/rule/some_kind_of_parser/]'
    """
    if len(sources) == 0:
        # Logically, a conjunction with 0 inputs represents 'True', a
        # proposition that cannot be denied. This could be useful as a
        # justification for, say, mathematical axioms, but when it comes to
        # ConceptNet, that kind of thing makes us uncomfortable and shouldn't
        # appear in the data.
        raise ValueError("Conjunctions of 0 things are not allowed")
    elif len(sources) == 1:
        return sources[0]
    else:
        return compound_uri("/and", sorted(set(sources)))


def assertion_uri(rel, start, end):
    """
    Make a URI for an assertion, as a compound URI of its relation, start node,
    and end node.

    >>> assertion_uri('/r/CapableOf', '/c/en/cat', '/c/en/sleep')
    '/a/[/r/CapableOf/,/c/en/cat/,/c/en/sleep/]'
    """
    assert rel.startswith("/r"), rel
    return compound_uri("/a", (rel, start, end))


def is_concept(uri):
    """
    >>> is_concept('/c/sv/klänning')
    True
    >>> is_concept('/x/en/ly')
    False
    >>> is_concept('/a/[/r/Synonym/,/c/ro/funcția_beta/,/c/en/beta_function/]')
    False
    """
    return uri.startswith("/c/")


def is_term(uri):
    """
    >>> is_term('/c/sv/kostym')
    True
    >>> is_term('/x/en/ify')
    True
    >>> is_term('/a/[/r/RelatedTo/,/c/en/cake/,/c/en/flavor/]')
    False
    """
    return uri.startswith("/c/") or uri.startswith("/x/")


def is_absolute_url(uri):
    """
    We have URLs pointing to Creative Commons licenses, starting with 'cc:', which for Linked
    Data purposes are absolute URLs because they'll be resolved into full URLs.
    >>> is_absolute_url('http://fr.wiktionary.org/wiki/mįkká’e_uxpáðe')
    True
    >>> is_absolute_url('/c/fr/nouveau')
    False
    """
    return uri.startswith("http") or uri.startswith("cc:")


def get_uri_language(uri):
    """
    Extract the language from a concept URI. If the URI points to an assertion,
    get the language of its first concept.
    >>> get_uri_language('/a/[/r/RelatedTo/,/c/en/orchestra/,/c/en/symphony/]')
    'en'
    >>> get_uri_language('/c/pl/cześć')
    'pl'
    >>> get_uri_language('/x/en/able')
    'en'
    """
    if uri.startswith("/a/"):
        return get_uri_language(parse_possible_compound_uri("a", uri)[1])
    elif is_term(uri):
        return split_uri(uri)[1]
    else:
        return None


def uri_to_label(uri):
    """
    Convert a ConceptNet uri into a label to be used in nodes. This
    function replaces an underscore with a space, so while '/c/en/example' will be converted into
    'example', '/c/en/canary_islands' will be converted into 'canary islands'.
    >>> uri_to_label('/c/en/example')
    'example'
    >>> uri_to_label('/c/en/canary_islands')
    'canary islands'
    >>> uri_to_label('/c/en')
    ''
    """
    if is_term(uri):
        uri = uri_prefix(uri)
    parts = split_uri(uri)
    if len(parts) < 3:
        return ""
    return parts[-1].replace("_", " ")


class Licenses:
    cc_attribution = "cc:by/4.0"
    cc_sharealike = "cc:by-sa/4.0"



import logging
from finntk.utils import ResourceMan, urlretrieve
from gensim.models import KeyedVectors
from shutil import copyfileobj
import os
import zipfile
from .base import MonolingualVectorSpace, RefType
from .utils import get

logger = logging.getLogger(__name__)


class Word2VecWordVecs(ResourceMan):
    RESOURCE_NAME = "word2vec-fi"

    URL = "http://vectors.nlpl.eu/repository/11/42.zip"

    def __init__(self):
        super().__init__()
        self._resources["vecs"] = "word2vec.binvec"
        self._vecs = None

    def _bootstrap(self, res):
        from gensim.test.utils import get_tmpfile

        logger.info("Downloading Word2Vec word vectors")
        zipped_tmp_fn = urlretrieve(self.URL)
        try:
            tmp_zip = zipfile.ZipFile(zipped_tmp_fn)
            tmp_fn = get_tmpfile("word2vec-fi.txt")
            try:
                copyfileobj(tmp_zip.open("model.txt"), open(tmp_fn, "wb"))
                logger.info("Converting Word2Vec word vectors")
                fi = KeyedVectors.load_word2vec_format(tmp_fn, unicode_errors="replace")
                fi.save(self._get_res_path("vecs"))
            finally:
                os.remove(tmp_fn)
        finally:
            os.remove(zipped_tmp_fn)

    def get_vecs(self):
        if self._vecs is None:
            vec_path = self.get_res("vecs")
            logger.info("Loading word vectors")
            self._vecs = KeyedVectors.load(vec_path, mmap="r")
            logger.info("Loaded word vectors")
        return self._vecs


vecs = Word2VecWordVecs()


class Word2VecFiSpace(MonolingualVectorSpace):
    takes = RefType.WORD_FORM
    dim = 100

    def get_vec(self, ref: str):
        return get(vecs.get_vecs(), ref)


space = Word2VecFiSpace()



from sklearn.neighbors import NearestNeighbors
import numpy as np
from os import makedirs
from os.path import join as pjoin, exists
import pickle


class WordExpert:

    def __init__(self):
        self.xs = []
        self.ys = []
        self.clf = None

    def add_word(self, ctx_vec, sense_key):
        vec_norm = np.linalg.norm(ctx_vec)
        self.xs.append(ctx_vec / vec_norm)
        self.ys.append(sense_key)

    def fit(self):
        clf = NearestNeighbors(n_neighbors=1, algorithm="ball_tree")
        clf.fit(self.xs, self.ys)
        self.clf = clf
        del self.xs

    def predict(self, ctx_vec):
        indices = self.clf.kneighbors([ctx_vec], return_distance=False)
        index = indices[0, 0]
        return self.ys[index]


class ExpertExists(Exception):
    pass


class WordExpertManager:

    def __init__(self, path, mode="r", allow_overwrites=False):
        self.path = path
        self.allow_overwrites = allow_overwrites
        if mode == "w":
            makedirs(path, exist_ok=allow_overwrites)

    def _expert_path(self, word):
        return pjoin(self.path, word)

    def dump_expert(self, word, expert):
        expert_path = self._expert_path(word)
        if not self.allow_overwrites and exists(expert_path):
            raise ExpertExists(f"Not overwriting existing word expert: {expert_path}")
        with open(expert_path, "wb") as outf:
            pickle.dump(expert, outf)

    def load_expert(self, word):
        expert_path = self._expert_path(word)
        if not exists(expert_path):
            return
        with open(expert_path, "rb") as inf:
            return pickle.load(inf)



"""
Functions for basic processing of OMorFi analyses.
"""
from more_itertools import split_at
import re
from itertools import product


def analysis_to_pairs(ana):
    for bit in ana.split("]["):
        k, v = bit.strip("[]").split("=", 1)
        yield k, v


def pairs_to_dict(it):
    return dict(((k.lower(), v) for k, v in it))


def analysis_to_dict(ana):
    return pairs_to_dict(analysis_to_pairs(ana))


def dict_to_analysis(d):
    return "[{}]".format(
        "][".join(["{}={}".format(k.upper(), v) for k, v in d.items()])
    )


def chunk_subwords(it):

    def is_cmp_bound(kv):
        return (kv[0] == "BOUNDARY" and kv[1] == "COMPOUND")

    return split_at(it, is_cmp_bound)


def analysis_to_subword_dicts(ana):
    """
    Returns a list of list of dicts. Each list element is an analysis. For each
    analysis, there is a list of subwords. Each dict contains an Omorfi
    analysis
    """
    return map(pairs_to_dict, chunk_subwords(analysis_to_pairs(ana)))


def generate_dict(ana):
    from .inst import get_omorfi

    omor = get_omorfi()
    ana_cp = ana.copy()
    if "weight" in ana_cp:
        del ana_cp["weight"]
    ana_txt = dict_to_analysis(ana_cp)
    return {gen["surf"] for gen in omor.generate(ana_txt)}


def generate_or_passthrough(ana):
    return {
        norm_word_id(ana["word_id"]) if s.startswith("[") else s
        for s in generate_dict(ana)
    }


def lemmas_of_subword_dicts(subword_dicts):
    subword_dicts = list(subword_dicts)
    return [
        "".join(prefixes) + norm_word_id(subword_dicts[-1]["word_id"])
        for prefixes in product(
            *(generate_or_passthrough(d) for d in subword_dicts[:-1])
        )
    ]


EXTRA_WORD_ID = re.compile(r"_\d+$")


def norm_word_id(word_id):
    extra_match = EXTRA_WORD_ID.search(word_id)
    if extra_match:
        word_id = word_id[:extra_match.start()]
    return word_id.lower()



import numpy as np
import logging
import os

logger = logging.getLogger(__name__)


def get_elmo(model_path, use_cuda=True):
    from elmoformanylangs.gen_elmo import Model
    from elmoformanylangs.modules.embedding_layer import EmbeddingLayer
    import codecs
    import json
    from os.path import dirname, join as pjoin
    config = json.load(open(pjoin(dirname(__file__), "../vendor/cnn_50_100_512_4096_sample.json")))
    # For the model trained with character-based word encoder.
    if config['token_embedder']['char_dim'] > 0:
        char_lexicon = {}
        with codecs.open(os.path.join(model_path, 'char.dic'), 'r', encoding='utf-8') as fpi:
            for line in fpi:
                tokens = line.strip().split('\t')
                if len(tokens) == 1:
                    tokens.insert(0, '\u3000')
                token, i = tokens
                char_lexicon[token] = int(i)
        char_emb_layer = EmbeddingLayer(config['token_embedder']['char_dim'], char_lexicon, fix_emb=False, embs=None)
        logging.info('char embedding size: ' + str(len(char_emb_layer.word2id)))
    else:
        char_lexicon = None
        char_emb_layer = None

    # For the model trained with word form word encoder.
    if config['token_embedder']['word_dim'] > 0:
        word_lexicon = {}
        with codecs.open(os.path.join(model_path, 'word.dic'), 'r', encoding='utf-8') as fpi:
            for line in fpi:
                tokens = line.strip().split('\t')
                if len(tokens) == 1:
                    tokens.insert(0, '\u3000')
                token, i = tokens
                word_lexicon[token] = int(i)
        word_emb_layer = EmbeddingLayer(config['token_embedder']['word_dim'], word_lexicon, fix_emb=False, embs=None)
        logging.info('word embedding size: ' + str(len(word_emb_layer.word2id)))
    else:
        word_lexicon = None
        word_emb_layer = None

    # instantiate the model
    model = Model(config, word_emb_layer, char_emb_layer, use_cuda)

    if use_cuda:
        model.cuda()

    logging.info(str(model))
    model.load_model(model_path)

    # I'll just keep this here...
    model.config = config
    model.char_lexicon = char_lexicon
    model.word_lexicon = word_lexicon
    model.use_cuda = use_cuda

    return model


def read_sent(sent, max_chars=None):
    dataset = []
    textset = []
    data = ['<bos>']
    text = []
    for token in sent:
        text.append(token)
        if max_chars is not None and len(token) + 2 > max_chars:
            token = token[:max_chars - 2]
        data.append(token)
    data.append('<eos>')
    dataset.append(data)
    textset.append(text)
    return dataset, textset


def get_layer(data, layer):
    if layer == -2:
        return np.r_[np.expand_dims(np.average(data, axis=0), 0), data]
    elif layer == -1:
        return np.average(data, axis=0)
    else:
        return data[layer]


def embed_sentence(model, sent, output_layer=-1):
    from elmoformanylangs.gen_elmo import create_one_batch
    config = model.config
    if config['token_embedder']['name'].lower() == 'cnn':
        batch_x, batch_text = read_sent(sent, config['token_embedder']['max_characters_per_token'])
    else:
        batch_x, batch_text = read_sent(sent)

    batch_w, batch_c, batch_lens, batch_masks = create_one_batch(
        batch_x,
        model.word_lexicon,
        model.char_lexicon,
        config,
        sort=False,
        use_cuda=model.use_cuda
    )

    output = model.forward(batch_w, batch_c, batch_masks)
    for i, text in enumerate(batch_text):
        if config['encoder']['name'].lower() == 'lstm':
            data = output[i, 1:batch_lens[i]-1, :].data
            if model.use_cuda:
                data = data.cpu()
            data = data.numpy()
        elif config['encoder']['name'].lower() == 'elmo':
            data = output[:, i, 1:batch_lens[i]-1, :].data
            if model.use_cuda:
                data = data.cpu()
            data = data.numpy()
    return get_layer(data, output_layer)


def read_sents(sents, max_chars=None):
    dataset = []
    textset = []
    for sent in sents:
        data = ['<bos>']
        text = []
        for token in sent:
            text.append(token)
            if max_chars is not None and len(token) + 2 > max_chars:
                token = token[:max_chars - 2]
            data.append(token)
        data.append('<eos>')
        dataset.append(data)
        textset.append(text)
    return dataset, textset


def create_batches(x, batch_size, word2id, char2id, config, use_cuda=False):
    from elmoformanylangs.gen_elmo import create_one_batch
    batches_w, batches_c, batches_lens, batches_masks = [], [], [], []
    size = batch_size
    nbatch = (len(x) - 1) // size + 1
    for i in range(nbatch):
        start_id, end_id = i * size, (i + 1) * size
        bw, bc, blens, bmasks = create_one_batch(
            x[start_id: end_id], word2id, char2id, config,
            sort=False, use_cuda=use_cuda)
        batches_w.append(bw)
        batches_c.append(bc)
        batches_lens.append(blens)
        batches_masks.append(bmasks)

    return batches_w, batches_c, batches_lens, batches_masks


def embed_sentences(model, sents, output_layer=-1, batch_size=64):
    config = model.config
    if config['token_embedder']['name'].lower() == 'cnn':
        batch_x, batch_text = read_sents(sents, config['token_embedder']['max_characters_per_token'])
    else:
        batch_x, batch_text = read_sents(sents)
    batch_w, batch_c, batch_lens, batch_masks = create_batches(
        batch_x, batch_size, model.word_lexicon, model.char_lexicon, config, use_cuda=model.use_cuda)
    res = []
    for w, c, lens, masks in zip(batch_w, batch_c, batch_lens, batch_masks):
        output = model.forward(w, c, masks)
        for i in range(len(lens)):
            if config['encoder']['name'].lower() == 'lstm':
                data = output[i, 1:lens[i]-1, :].data
            elif config['encoder']['name'].lower() == 'elmo':
                data = output[:, i, 1:lens[i]-1, :].data
            if model.use_cuda:
                data = data.cpu()
            data = data.numpy()
            res.append(get_layer(data, output_layer))
    return res



import logging
import os

logger = logging.getLogger(__name__)


class BertWordVecs:

    def __init__(self):
        super().__init__()
        self._model = None

    def get(self):
        from finntk.vendor.bert import get_bert

        if self._model is None:
            logger.info("Loading BERT")
            use_cuda = not os.environ.get("DISABLE_CUDA")
            self._model = get_bert("bert-base-multilingual-cased", use_cuda=use_cuda)
            logger.info("Loaded BERT")
        return self._model


vecs = BertWordVecs()
